import tensorflow as tf
from .common_functions.tf_fits_header import _hdu_condition, _hdu_body
from .table_functions.tf_fits_bintable_header import _read_TFORM_condition, _read_TFORM_body
from .table_functions.tf_fits_bintable_data import _bintable_column_condition, _bintable_column_body

@tf.function
def bintable_decode_fits(fits_data, header=1):
    """
    Function to decode fits binary table
    
    Parameters
    ----------
    fits_data : byte string
                byte string of data (from `tf.io.read_file(file_path)`)
    header : int
             header to return
    
    Returns
    -------
    table_data : tf.Tensor
                 tf.Tensor with dtype `tf.float32` of table data
    
    WARNING
    -------
    WILL MANGLE ANYTHING THAT IS NOT A PURE REAL NUMBER OR BOOL
    
    As Tensorflow wants a single data type returned, strings, complex numbers,
    arrays of numbers will be mangled."""
    
    h = tf.constant(0)
    offset = 0 #Position of start of HDU
    start = 0
    true_length = 0
    fixed_length = 0
    bitpix = 0
    shape = tf.TensorArray(dtype=tf.int32, size=0, dynamic_size=True, clear_after_read=False, name='shape')
    
    hdu_loop = tf.while_loop(_hdu_condition, _hdu_body,
                             [fits_data, h, header, offset, start, true_length, fixed_length, bitpix, shape],
                              shape_invariants=[None,
                                                h.get_shape(),
                                                h.get_shape(),
                                                h.get_shape(),
                                                h.get_shape(),
                                                h.get_shape(),
                                                h.get_shape(),
                                                h.get_shape(),
                                                tf.TensorShape(None)])
    _, _, _, offset, start, true_length, fixed_length, bitpix, shape = hdu_loop
    
    i = tf.constant(0)
    TFIELDS = 0
    TFORM = tf.TensorArray(dtype=tf.int32, size=0, dynamic_size=True, clear_after_read=False, name='TFORM')
    string = tf.strings.substr(fits_data, offset, 80)
    
    tform_loop = tf.while_loop(_read_TFORM_condition, _read_TFORM_body,
                               [fits_data, offset, i, string, TFIELDS, TFORM],
                               shape_invariants=[None,
                                                 h.get_shape(),
                                                 h.get_shape(),
                                                 None,
                                                 h.get_shape(),
                                                 None])
    _, _, _, _, TFIELDS, TFORM = tform_loop
    
    #Get btye data for the chosen header
    byte_data = tf.strings.substr(fits_data, offset+start, fixed_length)
    
    r = tf.constant(0)
    start = tf.constant(0)
    NAXIS1 = shape.read(0)
    NAXIS2 = shape.read(1)
    table_data = tf.TensorArray(dtype=tf.float32, size=0, dynamic_size=True, clear_after_read=False, name='table_data')
    
    col_loop = tf.while_loop(_bintable_column_condition, _bintable_column_body,
                             [r, NAXIS1, NAXIS2, TFIELDS, byte_data, start, fixed_length, TFORM, table_data],
                             shape_invariants=[r.get_shape(),
                                               NAXIS1.get_shape(),
                                               NAXIS2.get_shape(),
                                               TFIELDS.get_shape(),
                                               tf.TensorShape(None),
                                               tf.TensorShape(None),
                                               tf.TensorShape(None),
                                               tf.TensorShape(None),
                                               tf.TensorShape(None)],
                             name='col_loop')
    _, _, _, _, _, _, _, _, table_data = col_loop
    
    table_data = table_data.stack()
    table_data = tf.transpose(table_data)
    
    return table_data



import tensorflow as tf

#loop over columns
@tf.function
def _bintable_column_condition(r, NAXIS1, NAXIS2, TFIELDS, byte_data, start, fixed_length, TFORM, table_data):
    '''while loop condition when get data of column r
       r - colmn to get data
       NAXIS1 - number of bytes per row
       NAXIS2 - number of rows
       TFIELDS - noumber of columns
       byte_data - byte data of table being decoded
       start - start index of byte data
       fixed_length - length of byte_data
       TFORM - column butpix and number of type
       table_data = data of table
       
       returns if r < TFIELDS'''
    return tf.math.less(r, TFIELDS)

@tf.function
def _bintable_column_body(r, NAXIS1, NAXIS2, TFIELDS, byte_data, start, fixed_length, TFORM, table_data):
    '''Function to get data of column r
       r - colmn to get data
       NAXIS1 - number of bytes per row
       NAXIS2 - number of rows
       TFIELDS - noumber of columns
       byte_data - byte data of table being decoded
       start - start index of byte data
       fixed_length - length of byte_data
       TFORM - column butpix and number of type
       table_data = data of table'''
    
    #Get column data structure
    TFORMn = TFORM.read(r)
    j = tf.constant(0)
    bitpix = tf.slice(TFORMn, [0], [1])
    numtyp = tf.slice(TFORMn, [1], [1])
        
    #Get data through tf.strided_slice
    colm_data = tf.TensorArray(dtype=tf.float32, size=0, dynamic_size=True, name='colm_data')
    colm_start = start
    _, _, _, _, _, _, _, colm_data = tf.while_loop(_bintable_column_data_condition, _bintable_column_data_body,
                                                   [j, NAXIS1, TFIELDS, byte_data, colm_start, fixed_length, bitpix, colm_data],
                                                   shape_invariants=[j.get_shape(),
                                                                     j.get_shape(),
                                                                     tf.TensorShape(None),
                                                                     tf.TensorShape(None),
                                                                     colm_start.get_shape(),
                                                                     j.get_shape(),
                                                                     bitpix.get_shape(),
                                                                     tf.TensorShape(None)],
                                                   name='column_data')
    colm_data = colm_data.stack()
    colm_data = tf.reshape(colm_data, (NAXIS2,))
    
    table_data = table_data.write(r, colm_data)
    
    start += (numtyp*abs(bitpix))//8
    r += 1
    return r, NAXIS1, NAXIS2, TFIELDS, byte_data, start, fixed_length, TFORM, table_data

@tf.function
def _bintable_column_data_condition(j, NAXIS1, TFIELDS, byte_data, colm_start, fixed_length, bitpix, colm_data):
    '''while loop condition when get data of cell j of column r
       j - cell in row r
       NAXIS1 - number of bytes per row
       TFIELDS - noumber of columns
       byte_data - byte data of table being decoded
       colm_start - start index of cell
       fixed_length - length of byte_data
       bitpix - number of bytes in data type
       colm_data - Tensor of column data
       
       returns if colm_start < fixed_length'''
    return tf.math.less(colm_start, fixed_length)

@tf.function
def _bintable_column_data_body(j, NAXIS1, TFIELDS, byte_data, colm_start, fixed_length, bitpix, colm_data):
    '''Function to get data of cell j of column r
       j - cell in row r
       NAXIS1 - number of bytes per row
       TFIELDS - noumber of columns
       byte_data - byte data of table being decoded
       colm_start - start index of cell
       fixed_length - length of byte_data
       bitpix - number of bytes in data type
       colm_data - Tensor of column data'''
    
    length = tf.cast(abs(bitpix)//8, tf.int32)
    if tf.shape(tf.shape(colm_start))[0] == 1:
        colm_start = colm_start[0]
    if tf.shape(tf.shape(length))[0] == 1:
        length = length[0]
    
    #Decode cell data
    #Done cell-by-cell or the DMA string copy error occurs    
    cell_data = tf.strings.substr(byte_data, colm_start, length)
    if tf.math.equal(bitpix, 8):
        rcvd_data = tf.io.decode_raw(cell_data, tf.uint8, False)
        rcvd_data = tf.cast(rcvd_data, tf.float32)
    elif tf.math.equal(bitpix, 16):
        rcvd_data = tf.io.decode_raw(cell_data, tf.int16, False)
        rcvd_data = tf.cast(rcvd_data, tf.float32)
    elif tf.math.equal(bitpix, 32):
        rcvd_data = tf.io.decode_raw(cell_data, tf.int32, False)
        rcvd_data = tf.cast(rcvd_data, tf.float32)
    elif tf.math.equal(bitpix, 64):
        rcvd_data = tf.io.decode_raw(cell_data, tf.int64, False)
        rcvd_data = tf.cast(rcvd_data, tf.float32)
    elif tf.math.equal(bitpix, -32):
        rcvd_data = tf.io.decode_raw(cell_data, tf.float32, False)
    else:# tf.math.equal(bitpix, -64):
        rcvd_data = tf.io.decode_raw(cell_data, tf.float64, False)
        rcvd_data = tf.cast(rcvd_data, tf.float32)
    
    colm_data = colm_data.write(j, tf.reshape(rcvd_data, (1,)))
    colm_start += NAXIS1
    j += 1
    
    return j, NAXIS1, TFIELDS, byte_data, colm_start, fixed_length, bitpix, colm_data


import tensorflow as tf

#Get size and shape of data
@tf.function
def _size_and_shape_condition(axis, NAXIS, size, shape):
    '''while loop condition when get size and shape from NAXIS
       axis - current axis being used
       NAXIS - all axes sizes
       size - liner size of data
       shape - shape of final data
       
       returns if axis <= length of axes'''
    return tf.math.less_equal(axis, NAXIS.read(0))

@tf.function
def _size_and_shape_body(axis, NAXIS, size, shape):
    '''Function to get size and shape from NAXIS
       axis - current axis being used
       NAXIS - all axes sizes
       size - liner size of data
       shape - shape of final data'''
    
    NAXISn = NAXIS.read(axis)
    size *= NAXISn
    shape = shape.write(axis-1, NAXISn)
    axis += 1
    return axis, NAXIS, size, shape


import tensorflow as tf
from .common_functions.tf_fits_header import _hdu_condition, _hdu_body

@tf.function
def image_decode_fits(fits_data, header):
    """
    Function to decode fits binary table
    
    Parameters
    ----------
    fits_data : byte string
                byte string of data (from `tf.io.read_file(file_path)`)
    header : int
             header to return
    
    Returns
    -------
    img_data : tf.Tensor
               tf.Tensor with dtype `tf.float32` of image data
    """
    
    h = tf.constant(0)
    offset = 0 #Position of start of HDU
    start = 0
    true_length = 0
    fixed_length = 0
    bitpix = 0
    shape = tf.TensorArray(dtype=tf.int32, size=0, dynamic_size=True, clear_after_read=False, name='shape')
    
    hdu_loop = tf.while_loop(_hdu_condition, _hdu_body,
                             [fits_data, h, header, offset, start, true_length, fixed_length, bitpix, shape],
                              shape_invariants=[None,
                                                h.get_shape(),
                                                h.get_shape(),
                                                h.get_shape(),
                                                h.get_shape(),
                                                h.get_shape(),
                                                h.get_shape(),
                                                h.get_shape(),
                                                tf.TensorShape(None)])
    _, _, _, offset, start, true_length, fixed_length, bitpix, shape = hdu_loop
    
    #Get btye data for the chosen header
    byte_data = tf.strings.substr(fits_data, offset+start, true_length)
    #Decode the byte data
    if tf.math.equal(bitpix, 8):
        rcvd_data = tf.io.decode_raw(byte_data, tf.int8, False, fixed_length)
        rcvd_data = tf.cast(rcvd_data, tf.float32)
    elif tf.math.equal(bitpix, 16):
        rcvd_data = tf.io.decode_raw(byte_data, tf.int16, False, fixed_length)
        rcvd_data = tf.cast(rcvd_data, tf.float32)
    elif tf.math.equal(bitpix, 32):
        rcvd_data = tf.io.decode_raw(byte_data, tf.int32, False, fixed_length)
        rcvd_data = tf.cast(rcvd_data, tf.float32)
    elif tf.math.equal(bitpix, 64):
        rcvd_data = tf.io.decode_raw(byte_data, tf.int64, False, fixed_length)
        rcvd_data = tf.cast(rcvd_data, tf.float32)
    elif tf.math.equal(bitpix, -32):
        rcvd_data = tf.io.decode_raw(byte_data, tf.float32, False, fixed_length)
    else:# tf.math.equal(bitpix, -64):
        rcvd_data = tf.io.decode_raw(byte_data, tf.float64, False, fixed_length)
        rcvd_data = tf.cast(rcvd_data, tf.float32)
        
    #Swap  the first two axes
    ax0 = shape.read(0)
    ax1 = shape.read(1)
    shape = shape.write(0, ax1)
    shape = shape.write(1, ax0)
    
    #Reshape the data
    shape = shape.stack()
    img_data = tf.reshape(rcvd_data, shape)
    
    return img_data


import tensorflow as tf
from .tf_fits_common import _size_and_shape_condition, _size_and_shape_body

#Loop through Header Keywords
@tf.function
def _read_header_condition(fits_data, NAXIS, bitpix, offset, i, string):
    '''while loop condition when read next keyword in header
       and get necessary data from the keyword
       fits_data - byte string of data (from tf.io.read_file(file_path))
       NAXIS - all axes sizes
       TFIELDS - number of columns in bintable
       bitpix - fits data type
       offset - end of last header
       i - current keyword
       string - the keyword string
       TFORM - table column data type strings
       
       returns if keyword is not the final keyword'''
    return tf.math.not_equal(string, b'END                                                                             ')

@tf.function
def _read_header_body(fits_data, NAXIS, bitpix, offset, i, string):
    '''Function to read next keyword in header
       and get necessary data from the keyword
       fits_data - byte string of data (from tf.io.read_file(file_path))
       NAXIS - all axes sizes
       bitpix - fits data type
       offset - end of last header
       i - current keyword
       string - the keyword string'''
    
    #Get next keyword
    #All keyword lines are 80 characters long
    string = tf.strings.substr(fits_data, offset+(i*80), 80)
    
    #NAXIS - get number of axes
    if tf.math.equal(b'NAXIS   =', tf.strings.substr(string, 0, 9)):
        NAXIS = NAXIS.write(0, int(tf.strings.substr(string, 9, 21)))
        
    #NAXIS# - get size of axis #
    elif tf.math.equal(b'NAXIS', tf.strings.substr(string, 0, 5)):
        axis = int(tf.strings.substr(string, 5, 3))
        NAXIS = NAXIS.write(axis, int(tf.strings.substr(string, 9, 21)))
        
    #BITPIX - get bit size of values
    elif tf.math.equal(b'BITPIX  =', tf.strings.substr(string, 0, 9)):
        bitpix = int(tf.strings.substr(string, 9, 21))
        
    i += 1
    return fits_data, NAXIS, bitpix, offset, i, string

#Loop through HDUs
@tf.function
def _hdu_condition(fits_data, h, header, offset, start, true_length, fixed_length, bitpix, shape):
    '''while loop condition when read header of HDU h
       called in tf.while_loop
       fits_data - byte string of data (from tf.io.read_file(file_path))
       h - current header
       header - target header
       offset - end of last header
       start - start of data
       true_length - length of data (including padding)
       fixed_length - length of data (excluding padding)
       bitpix - fits data type
       shape - shape of data
       
       returns if h <= header'''
    return tf.math.less_equal(h, header)
    
@tf.function
def _hdu_body(fits_data, h, header, offset, start, true_length, fixed_length, bitpix, shape):
    '''Function to read header of HDU h
       called in tf.while_loop
       fits_data - byte string of data (from tf.io.read_file(file_path))
       h - current header
       header - target header
       offset - end of last header
       start - start of data
       true_length - length of data (including padding)
       fixed_length - length of data (excluding padding)
       bitpix - fits data type
       shape - shape of data'''
    
    #All header keywords are 80 characters long
    i0 = tf.constant(0)
    i = 0
    NAXIS = tf.TensorArray(dtype=tf.int32, size=0, dynamic_size=True, clear_after_read=False, name='NAXIS')
    string = tf.strings.substr(fits_data, offset, 80)
    
    read_loop = tf.while_loop(_read_header_condition, _read_header_body,
                              [fits_data, NAXIS, bitpix, offset, i, string],
                              shape_invariants=[None,
                                                None,
                                                i0.get_shape(),
                                                i0.get_shape(),
                                                i0.get_shape(),
                                                None],
                              name='read')
    _, NAXIS, bitpix, offset, i, _ = read_loop

    i -= 1
    start = i*80
        
    #All data blocks are 2880 bytes long,
    #so find where the data starts including padding
    #For header, padding is spaces: b' '
    tru_start = start // 2880
    if start % 2880 != 0:
        tru_start += 1
    start = tru_start*2880

    #Get bytes per value
    out_size = 0
    if tf.math.equal(bitpix, 8):
        out_size = 1
    elif tf.math.equal(bitpix, 16):
        out_size = 2
    elif tf.math.equal(bitpix, 32):
        out_size = 4
    elif tf.math.equal(bitpix, 64):
        out_size = 8
    elif tf.math.equal(bitpix, -32):
        out_size = 4
    elif tf.math.equal(bitpix, -64):
        out_size = 8

    #Get byte size of data
    #and shape of the final data
    if tf.math.greater(NAXIS.read(0), 0):
        size = 1
        axis = tf.constant(1)
        axis, NAXIS, size, shape = tf.while_loop(_size_and_shape_condition, _size_and_shape_body,
                                                 [axis, NAXIS, size, shape],
                                                 shape_invariants=[axis.get_shape(),
                                                                   None,
                                                                   axis.get_shape(),
                                                                   None],
                                                 name='size_and_shape')                                 
        fixed_length = size*out_size
    else:
        fixed_length = 0

    #All data blocks are 2880 bytes long,
    #so find how long the data is including pading
    #For data, padding is zeros: b'\x00'
    true_length = int(fixed_length//2880)
    if tf.math.not_equal(fixed_length % 2880, 0):
        true_length += 1
    true_length *= 2880
    
    if tf.math.not_equal(h, header):
        #If not the HDU we want,
        #increment offset by HDU size
        offset += start+true_length
        
    h += 1
    
    return fits_data, h, header, offset, start, true_length, fixed_length, bitpix, shape



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tf-fits",
    version="1.0.1",
    author="Willaim J. Pearson",
    author_email="willjamespearson@gmail.com",
    license="Apache-2.0",
    description="Load FITS files into tf.data.Dataset",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wjpearson/tensorflow_fits",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
                ],
    python_requires='>=3.5',
    install_requires=['tensorflow>=2.0.0a0'],
)



import tensorflow as tf

@tf.function
def _TFORM_bitpix(TFORMn):
    '''Function to get TFORM data type
       TFORMn - TFORM data type character
       
       returns bitpix - number of bits for data type'''  
    
    bitpix = 0
    if tf.strings.regex_full_match(TFORMn, b'L'):
        bitpix = 8
    elif tf.strings.regex_full_match(TFORMn, b'X'):
        bitpix = 0
    elif tf.strings.regex_full_match(TFORMn, b'B'):
        bitpix = 8
    elif tf.strings.regex_full_match(TFORMn, b'I'):
        bitpix = 16
    elif tf.strings.regex_full_match(TFORMn, b'J'):
        bitpix = 32
    elif tf.strings.regex_full_match(TFORMn, b'K'):
        bitpix = 64
    elif tf.strings.regex_full_match(TFORMn, b'A'):
        bitpix = 8
    elif tf.strings.regex_full_match(TFORMn, b'E'):
        bitpix = -32
    elif tf.strings.regex_full_match(TFORMn, b'D'):
        bitpix = -64
    elif tf.strings.regex_full_match(TFORMn, b'C'):
        bitpix = -64
    elif tf.strings.regex_full_match(TFORMn, b'M'):
        bitpix = -128
    elif tf.strings.regex_full_match(TFORMn, b'P'):
        bitpix = 32
    else:# tf.strings.regex_full_match(TFORMn, b'Q'):
        bitpix = 64
        
    return bitpix

#Loop through Header Keywords
@tf.function
def _read_TFORM_condition(fits_data, offset, i, string, TFIELDS, TFORM):
    '''while loop condition when read next keyword in header
       and get necessary data from the keyword
       fits_data - byte string of data (from tf.io.read_file(file_path))
       NAXIS - all axes sizes
       TFIELDS - number of columns in bintable
       bitpix - fits data type
       offset - end of last header
       i - current keyword
       string - the keyword string
       TFORM - table column data type strings
       
       returns if keyword is not the final keyword'''
    return tf.math.not_equal(string, b'END                                                                             ')

@tf.function
def _read_TFORM_body(fits_data, offset, i, string, TFIELDS, TFORM):
    '''Function to read next keyword in header
       and get necessary data from the keyword
       fits_data - byte string of data (from tf.io.read_file(file_path))
       offset - end of last header
       i - current keyword
       string - the keyword string
       TFIELDS - number of columns in bintable
       TFORM - table column data type strings'''
    
    #Get next keyword
    #All keyword lines are 80 characters long
    string = tf.strings.substr(fits_data, offset+(i*80), 80)
        
    #TFIELDS - get number of fields
    if tf.math.equal(b'TFIELDS =', tf.strings.substr(string, 0, 9)):
        TFIELDS = int(tf.strings.substr(string, 9, 21))
        
    #TFORM# - get field # data format
    elif tf.math.equal(b'TFORM', tf.strings.substr(string, 0, 5)):
        field = int(tf.strings.substr(string, 5, 3))-1
        TFORMn = tf.strings.substr(string, 11, 8)
        TFORMn = tf.strings.strip(TFORMn)
        
        len_TFORMn = tf.strings.length(TFORMn)
        num_TFORMn = 1
        if tf.math.greater(len_TFORMn, 1):
            num_TFORMn = int(tf.strings.substr(TFORMn, 0, len_TFORMn-1))
            TFORMn = tf.strings.substr(TFORMn, len_TFORMn-1, 1)
        bitpix_TFORMn = _TFORM_bitpix(TFORMn)
        
        TFORM = TFORM.write(field, [bitpix_TFORMn,num_TFORMn])
        
    i += 1
    return fits_data, offset, i, string, TFIELDS, TFORM